import xml.etree.ElementTree as ET
from os.path import splitext
from urllib.parse import urlsplit

import requests

from ..utils.urls import validate_url


def request_sitemap(url):
    """Retrieve the sitemap from the URL provided in cmd args."""

    url = validate_url(url)

    if validate_sitemap_url(url) is not True:
        err = "Invalid sitemap provided. Please provide a link to a valid XML sitemap."
        raise SystemExit(err)
    try:
        print(f"Requesting sitemap... ({url})")
        resp = requests.get(url)
        resp.raise_for_status()
    except requests.exceptions.HTTPError as errh:
        raise SystemExit(errh)
    except requests.exceptions.ConnectionError as errc:
        raise SystemExit(errc)
    except requests.exceptions.Timeout as errt:
        raise SystemExit(errt)
    except requests.exceptions.RequestException as err:
        raise SystemExit(err)

    sitemap = resp.text
    print("Sitemap retrieval successful!")

    return sitemap


def parse_sitemap(sitemap):
    """Parse URLs from the XML sitemap and return a list of URLs."""

    print("Parsing URLs from sitemap...")

    root = ET.fromstring(sitemap)
    namespace = "{http://www.sitemaps.org/schemas/sitemap/0.9}"

    urls = []
    for url in root.findall(f"{namespace}url"):
        loc = url.find(f"{namespace}loc")
        urls.append(loc.text)

    return urls


def validate_sitemap_url(url):
    """Validate that the sitemap URL is valid (.xml format)."""

    u = urlsplit(url)
    ext = splitext(u.path)[1]
    if ext == ".xml":
        return True



import copy
import json
from datetime import datetime

from ..cli.choices import COMMAND_CHOICES


def process_json(json_resp, category, strategy):
    """Dump raw json response to a file in the working directory."""

    date = _get_timestamp(json_resp)
    filename = f"psi-s-{strategy}-c-{category}-{date}.json"

    with open(filename, "w", encoding="utf-8") as f:
        json.dump(json_resp, f, ensure_ascii=False, indent=4)
        print("JSON processed. Check your current directory.")


def process_excel(json_resp, category, metrics):
    """Call various parsing operations for Excel / Sitemap formats."""

    audits_base = _get_audits_base(json_resp)  # Location of audits in json response

    metadata = _parse_metadata(json_resp, category)
    audit_results = _parse_audits(audits_base)

    # Metrics are only available in the performance category.
    if metrics is not None and category == "performance":
        metrics_results = _parse_metrics(audits_base, metrics)
    else:
        metrics_results = None

    results = {
        "metadata": metadata,
        "audit_results": audit_results,
        "metrics_results": metrics_results,
    }

    return results


def _parse_metadata(json_resp, category):
    """Parse various metadata from the JSON response to write to Excel."""

    json_base = json_resp["lighthouseResult"]
    strategy = json_base["configSettings"]["formFactor"]
    category_score = json_base["categories"][category]["score"]
    timestamp = _get_timestamp(json_resp)

    metadata = {
        "category": category,
        "category_score": category_score,
        "strategy": strategy,
        "timestamp": timestamp,
    }

    return metadata


def _parse_audits(audits_base):
    """Parse Lighthouse audits from the JSON response to write to Excel."""

    audit_results = {}

    # Create results dict with scores and numerical values for each audit.
    for k in audits_base.keys():
        score = audits_base[k].get("score")
        if score is not None:
            num_value = audits_base[k].get("numericValue", "n/a")
            audit_results[k] = [score * 100, num_value]
        else:
            audit_results[k] = ["n/a", "n/a"]

    # Sort dict alphabetically so each audit is written to Excel in the same order.
    audit_results = dict(sorted(audit_results.items()))

    return audit_results


def _parse_metrics(audits_base, metrics):
    """Parse performance metrics from the JSON response to write to Excel."""

    metrics_results = {}
    metrics_loc = audits_base["metrics"]["details"]["items"][0]

    if "all" in metrics:
        metrics_to_use = copy.copy(COMMAND_CHOICES["metrics"])
        # Remove 'all' to avoid key errors, as it doesn't exist in JSON resp.
        metrics_to_use.remove("all")
    else:
        metrics_to_use = metrics

    # Create new dict of metrics based on user's chosen metrics.
    for field in metrics_to_use:
        metric = metrics_loc[field]
        metrics_results[field] = metric

    # Sort dict alphabetically so each metric is written to Excel in the same order.
    metrics_results = dict(sorted(metrics_results.items()))

    return metrics_results


def _get_audits_base(json_resp):
    """Return the location of audits in the JSON response."""

    return json_resp["lighthouseResult"]["audits"]


def _get_timestamp(json_resp):
    """
    Parse the timestamp of the analysis from the JSON response.

    Covert it to a Python datetime object.
    """

    timestamp = json_resp["analysisUTCTimestamp"]
    date = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%fZ")

    return date



import asyncio

import aiohttp

from ..utils.urls import validate_url
from .keys import get_api_key

next_delay = 1  # Global for applying a 1s delay between requests


async def get_response(
    url,
    category=None,
    locale=None,
    strategy=None,
    utm_campaign=None,
    utm_source=None,
    captcha_token=None,
):
    """
    Make async GET calls to the PSI API for the requested page's URL.

    Return the awaited json response from the server.
    """

    base_url = "https://www.googleapis.com/pagespeedonline/v5/runPagespeed"
    url = validate_url(url)
    params = {
        "url": url,
        "category": category,
        "locale": locale,
        "strategy": strategy,
        "utm_campaign": utm_campaign,
        "utm_source": utm_source,
        "captcha_token": captcha_token,
    }
    # Use API defaults instead of passing None values as query params.
    params = {k: v for k, v in params.items() if v is not None}

    params["key"] = get_api_key()

    # Add a 1s delay between calls to avoid 500 errors from server.
    global next_delay
    next_delay += 1
    await asyncio.sleep(next_delay)

    print(f"Sending request... ({params['url']})")

    # Make async call with query params to PSI API and await response.
    async with aiohttp.ClientSession() as session:
        async with session.get(url=base_url, params=params) as resp:

            # Retry on errors up to 5 times. Sleep for 1s between retries.
            json_resp = None
            retry_attempts = 5
            while json_resp is None:
                try:
                    resp.raise_for_status()
                    json_resp = await resp.json()
                    print(f"Request successful! ({params['url']})")
                except aiohttp.ClientError as err_c:
                    if retry_attempts < 1:
                        raise aiohttp.ClientError(err_c)
                    else:
                        print(err_c)
                        retry_attempts -= 1
                        await asyncio.sleep(1)

    return json_resp


async def gather_responses(request_urls, api_args_dict):
    """Gather tasks and await the return of the responses for processing."""

    tasks = get_tasks(request_urls, api_args_dict)
    print(f"Preparing {len(tasks)} URL(s)...")

    responses = await asyncio.gather(*tasks)
    print(f"{len(responses)}/{len(tasks)} URL(s) processed successfully.")

    return responses


def get_tasks(request_urls, api_args_dict):
    """Create a list of tasks that call get_response() with request params."""

    tasks = []
    for url in request_urls:
        api_args_dict["url"] = url
        tasks.append(get_response(**api_args_dict))

    return tasks



"""
Define argpase command line argument choices.

'category', 'strategy' and 'locale' are all based on available PSI API query params.

'format' is specific to pyspeedinsights and dictates the output format of the results.

'metrics' defines available performance metrics to include. 'all' = collect all metrics.
"""

COMMAND_CHOICES = {
    "category": ["accessibility", "best-practices", "performance", "pwa", "seo"],
    "strategy": ["desktop", "mobile"],
    "locale": [
        "ar",
        "bg",
        "ca",
        "zh-TW",
        "zh-CN",
        "hr",
        "cs",
        "da",
        "nl",
        "en",
        "en-GB",
        "fil",
        "fi",
        "fr",
        "de",
        "el",
        "iw",
        "hi",
        "hu",
        "id",
        "it",
        "ja",
        "ko",
        "lv",
        "lt",
        "no",
        "pl",
        "pt-BR",
        "pt-PT",
        "ro",
        "ru",
        "sr",
        "sk",
        "sl",
        "es",
        "sv",
        "th",
        "tr",
        "uk",
        "vi",
    ],
    "format": ["json", "excel", "sitemap"],
    # `all` is used as an arg to include all metrics in the Excel output
    "metrics": [
        "all",
        "observedTotalCumulativeLayoutShift",
        "observedCumulativeLayoutShift",
        "observedLargestContentfulPaintAllFrames",
        "maxPotentialFID",
        "observedSpeedIndexTs",
        "observedFirstContentfulPaintTs",
        "observedTimeOrigin",
        "observedFirstPaint",
        "observedNavigationStartTs",
        "observedLargestContentfulPaintAllFramesTs",
        "speedIndex",
        "observedFirstContentfulPaint",
        "observedLastVisualChangeTs",
        "cumulativeLayoutShiftMainFrame",
        "observedLastVisualChange",
        "cumulativeLayoutShift",
        "largestContentfulPaint",
        "observedDomContentLoaded",
        "firstContentfulPaint",
        "observedCumulativeLayoutShiftMainFrame",
        "observedFirstVisualChange",
        "observedFirstPaintTs",
        "totalCumulativeLayoutShift",
        "observedFirstMeaningfulPaint",
        "interactive",
        "observedTraceEnd",
        "observedFirstMeaningfulPaintTs",
        "totalBlockingTime",
        "observedFirstContentfulPaintAllFramesTs",
        "observedLargestContentfulPaint",
        "observedNavigationStart",
        "observedLoad",
        "observedFirstVisualChangeTs",
        "observedFirstContentfulPaintAllFrames",
        "observedTimeOriginTs",
        "observedTraceEndTs",
        "observedLoadTs",
        "observedDomContentLoadedTs",
        "observedSpeedIndex",
        "firstMeaningfulPaint",
        "observedLargestContentfulPaintTs",
    ],
}



from dataclasses import dataclass, field

import xlsxwriter


@dataclass
class ExcelWorkbook:
    """Class for creating an Excel Workbook and writing the PSI API results to it."""

    url: str
    metadata: dict
    audit_results: dict
    metrics_results: dict = None
    workbook: xlsxwriter.Workbook = None
    worksheet: xlsxwriter.Workbook.worksheet_class = None
    cur_cell: list[int] = None
    category_scores: list[int] = field(default_factory=list)

    def set_up_worksheet(self):
        """Create the workbook, add a worksheet, and set up column headings."""

        self._create_workbook()
        self.worksheet = self.workbook.add_worksheet()

        row, col = 0, 0  # First cell
        column_format = self._column_format()
        metadata_format = self._metadata_format()

        # Add metadata to the first cell of the Excel sheet.
        category = self.metadata["category"].upper()
        strategy = self.metadata["strategy"].upper()
        metadata_value = f"{strategy} {category} REPORT"
        self.worksheet.write(row, col, metadata_value, metadata_format)

        # Add the URL heading with a wider column of merged cells.
        row += 2
        col += 1
        self.worksheet.set_column(col, col, 15)
        self.worksheet.merge_range(row, col, row, col + 3, "URL", column_format)

        # Add a column heading for each page's overall category score.
        col += 4
        self.worksheet.write(row, col, "Ovr", column_format)

        # Set the current cell in [row, col] format for later use.
        self.cur_cell = [row, col]

    def write_to_worksheet(self, is_first):
        """Write the URL, overall page score, audit and metrics results to the sheet."""

        if self.worksheet is not None:
            # Write the results to the Excel sheet.
            self._write_page_url()
            self._write_overall_category_score()
            self._write_audit_results(self.audit_results, is_first)

            if self.metrics_results is not None:
                self._write_metrics_results(self.metrics_results, is_first)

            # Move one row down for the next page and reset the column.
            self.cur_cell[0] += 1
            self.cur_cell[1] = 5
        else:
            raise ValueError("The worksheet is not set up. ")

    def finalize_and_save(self):
        """Write the average score to the workbook and save/close it."""

        self._write_average_score()

        self.workbook.close()
        print("Workbook saved. Check your current directory!")

    def _create_workbook(self):
        """Create an Excel workbook with a unique and descriptive name."""

        strategy = self.metadata["strategy"]
        category = self.metadata["category"]
        date = self.metadata["timestamp"]

        self.workbook = xlsxwriter.Workbook(
            f"psi-s-{strategy}-c-{category}-{date}.xlsx"
        )

    def _write_overall_category_score(self):
        """Write the ovr category score for the page / URL to the sheet."""

        category_score = self.metadata["category_score"] * 100
        cat_score_format = self._score_format(category_score)

        self.worksheet.write(
            self.cur_cell[0] + 2, self.cur_cell[1], category_score, cat_score_format
        )

        self.cur_cell[1] += 2

        # Add the score to a list so they can all be averaged at the end.
        self.category_scores.append(category_score)

    def _write_average_score(self):
        """Write the average score next to the worksheet metadata."""

        scores = self.category_scores
        avg_score = sum(scores) / len(scores)
        avg_score = round(avg_score, 2)

        format = self._metadata_format()
        self.worksheet.write(0, 4, f"Avg Score: {avg_score}", format)

    def _write_page_url(self):
        """
        Write the requested page URL being analyzed to the sheet.

        self.cur_cell: Row 2, Col 5
        """
        url_format = self._url_format()

        row = self.cur_cell[0] + 2

        self.worksheet.merge_range(
            row, 0, row, self.cur_cell[1] - 1, self.url, url_format
        )

    def _write_audit_results(self, audit_results, is_first):
        """Iterate through the audit results and write them to the worksheet."""

        column_format = self._column_format()
        row, col = self.cur_cell[0], self.cur_cell[1]

        for k, v in audit_results.items():
            self.worksheet.set_column(col, col + 1, 15)

            score = v[0]
            value = v[1]
            score_format = self._score_format(score)

            if is_first:
                self._write_results_headings(
                    row, col, k, column_format, result_type="audit"
                )
            self.worksheet.write(row + 2, col, score, score_format)
            self.worksheet.write(row + 2, col + 1, value, score_format)

            col += 2

        self.cur_cell[1] = col + 2

    def _write_metrics_results(self, metrics_results, is_first):
        """Iterate through the metrics results and write them to the worksheet."""

        column_format = self._column_format()
        data_format = self._data_format()
        row, col = self.cur_cell[0], self.cur_cell[1]

        for k, v in metrics_results.items():
            self.worksheet.set_column(col, col, 30)

            if is_first:
                self._write_results_headings(
                    row, col, k, column_format, result_type="metrics"
                )
            self.worksheet.write(row + 2, col, v, data_format)
            col += 1

    def _write_results_headings(self, row, col, k, column_format, result_type):
        """Write the headings for audit or metrics results to the worksheet."""

        if result_type == "audit":
            self.worksheet.merge_range(row, col, row, col + 1, k, column_format)
            self.worksheet.write(row + 1, col, "Score", column_format)
            self.worksheet.write(row + 1, col + 1, "Value", column_format)
        elif result_type == "metrics":
            self.worksheet.write(row, col, k, column_format)
            self.worksheet.write(row + 1, col, "Value", column_format)

    def _column_format(self):
        """Reusable formatting for column cells."""

        return self.workbook.add_format(
            {"font_size": 16, "bold": 1, "align": "center", "valign": "vcenter"}
        )

    def _data_format(self):
        """Reusable formatting for cells with regular data."""

        return self.workbook.add_format(
            {"font_size": 14, "align": "center", "valign": "vcenter"}
        )

    def _url_format(self):
        """Reusable formatting for cells with URLs."""

        return self.workbook.add_format(
            {"font_size": 14, "align": "left", "valign": "vcenter"}
        )

    def _metadata_format(self):
        """Reusable formatting for cells with metadata."""

        return self.workbook.add_format(
            {"font_size": 20, "bold": 1, "align": "left", "valign": "vcenter"}
        )

    def _score_format(self, score):
        """Reusable formatting and color coding for cells with scores."""

        if score == "n/a":
            color = "white"
        elif score >= 90:
            color = "lime"
        elif 90 > score >= 80:
            color = "green"
        elif 80 > score >= 70:
            color = "yellow"
        elif 70 > score >= 60:
            color = "orange"
        elif 60 > score >= 50:
            color = "brown"
        elif 50 > score:
            color = "red"

        return self.workbook.add_format(
            {"bg_color": color, "font_size": 14, "align": "center", "valign": "vcenter"}
        )



import argparse

from .choices import COMMAND_CHOICES


def set_up_arg_parser():
    """
    Setup argument parser with grouped command line arguments.

    Used to define PSI API request query parameters and how responses are processed.
    """

    parser = argparse.ArgumentParser(prog="pyspeedinsights")

    # Add argument options for default API call query params.
    api_group = parser.add_argument_group("API Group")

    api_group.add_argument(
        "url", help="The URL or Sitemap URL of the site you want to analyze."
    )
    api_group.add_argument(
        "-c",
        "--category",
        metavar="\b",
        dest="category",
        choices=COMMAND_CHOICES["category"],
        help="The Lighthouse category to run:\
            accessibility, best-practices, performance (default), pwa or seo.",
    )
    api_group.add_argument(
        "-l",
        "--locale",
        metavar="\b",
        dest="locale",
        choices=COMMAND_CHOICES["locale"],
        help="The locale used to localize formatted results.\
            Defaults to English (US).",
    )
    api_group.add_argument(
        "-s",
        "--strategy",
        metavar="\b",
        dest="strategy",
        choices=COMMAND_CHOICES["strategy"],
        help="The analysis strategy to use: desktop (default) or mobile.",
    )
    api_group.add_argument(
        "-uc",
        "--campaign",
        metavar="\b",
        dest="utm_campaign",
        help="The UTM campaign name for analytics.",
    )
    api_group.add_argument(
        "-us",
        "--source",
        metavar="\b",
        dest="utm_source",
        help="The UTM campaign source for analytics.",
    )
    api_group.add_argument(
        "-t",
        "--token",
        metavar="\b",
        dest="captcha_token",
        help="The captcha token passed when filling out a captcha.",
    )

    # Add other argument options for how to process the API response.
    proc_group = parser.add_argument_group("Processing Group")

    proc_group.add_argument(
        "-f",
        "--format",
        metavar="\b",
        dest="format",
        choices=COMMAND_CHOICES["format"],
        help="The format of the results: json (default), excel or sitemap.\
            json outputs all response data to a json file (1 URL at a time).\
            excel writes Lighthouse audits and (optionally) metrics\
                to an Excel file (1 URL at a time).\
            sitemap parses the sitemap URL you provide and\
                collects data for all your pages to Excel.",
    )
    proc_group.add_argument(
        "-m",
        "--metrics",
        metavar="\b",
        dest="metrics",
        choices=COMMAND_CHOICES["metrics"],
        nargs="+",
        help="The additional metric(s) to include in your report.\
            For Excel format only (the json output includes all metrics).\
            If excluded, only the default Lighthouse audits\
                will be saved to Excel.\
            Add the `all` argument to retrieve all available metrics.",
    )

    return parser


def parse_args(parser):
    """
    Parse user arguments from the command line.

    With `psi` console script entry point or when invoking the module directly.
    """

    args = parser.parse_args()
    return args


def create_arg_groups(parser, args):
    """
    Create separate namespaces for each arg group.

    Necessary for passing each arg group separately as kwargs.
    """

    arg_groups = {}

    for group in parser._action_groups:
        group_dict = {a.dest: getattr(args, a.dest, None) for a in group._group_actions}
        arg_groups[group.title] = argparse.Namespace(**group_dict)

    return arg_groups



from urllib.parse import urlsplit


def validate_url(url):
    """
    Adds a scheme to the URL if missing and
    validates that the URL is fully qualified (not just a path).
    """

    err = "Invalid URL. Please enter a valid Fully-Qualified Domain Name (FQDN)."
    u = urlsplit(url)

    if not (u.scheme and u.netloc):
        if "." not in u.path:
            raise SystemExit(err)
        u = u._replace(scheme="https", netloc=u.path, path="")

    return u.geturl()



import keyring


def get_api_key():
    """
    Get the user's PSI API key from their keystore using keyring.

    Exit the program if key is not set.
    """

    PSI_API_KEY = keyring.get_password("system", "psikey")

    if PSI_API_KEY is not None:
        return PSI_API_KEY
    else:
        err = "Error: Your PageSpeed Insights API key is empty.\
              \nGenerate a key with Google and set it with the command\
                  `keyring set system psikey`.\
              \nTo verify your key can be found, run the command\
                  `keyring get system psikey`."
        raise SystemExit(err)



import asyncio

from .api.request import gather_responses
from .api.response import process_excel, process_json
from .cli import commands
from .core.excel import ExcelWorkbook
from .core.sitemap import parse_sitemap, request_sitemap


def main():
    """
    Main execution loop of the program.

    Run with `psi` console script entry point or when invoking the module directly.
    """
    # Set up arg parser and parse cmd line args into groups.
    parser = commands.set_up_arg_parser()
    args = commands.parse_args(parser)
    arg_groups = commands.create_arg_groups(parser, args)

    # Convert arg groups to dicts.
    api_args_dict = vars(arg_groups["API Group"])
    proc_args_dict = vars(arg_groups["Processing Group"])

    # Remove keys with None values to avoid overriding kwargs.
    proc_args_dict = {k: v for k, v in proc_args_dict.items() if v is not None}

    format = proc_args_dict.get("format")
    metrics = proc_args_dict.get("metrics")

    url = api_args_dict.get("url")
    category = api_args_dict.get("category")
    strategy = api_args_dict.get("strategy")

    # API's default category and strategy with no query params.
    if category is None:
        category = "performance"
    if strategy is None:
        strategy = "desktop"

    if format == "sitemap":
        # Create list of request URLs based on sitemap.
        sitemap_url = url
        sitemap = request_sitemap(sitemap_url)
        request_urls = parse_sitemap(sitemap)
        request_urls = list(set(request_urls))  # Remove duplicates if they exist.
    else:
        # For analyzing a single page, only process the requested URL.
        request_urls = [url]

    # Make async requests to PSI API and gather responses.
    responses = asyncio.run(gather_responses(request_urls, api_args_dict))

    json_output = format == "json" or format is None
    excel_output = format in ["excel", "sitemap"]

    # Process and write data for each response based on chosen format.
    for i, response in enumerate(responses):

        if json_output:
            # Output raw JSON response to the current directory.
            process_json(response, category, strategy)

        elif excel_output:
            # Parse metadata, audit and metrics results for Excel.
            excel_results = process_excel(response, category, metrics)

            metadata = excel_results["metadata"]
            audit_results = excel_results["audit_results"]
            metrics_results = excel_results["metrics_results"]
            final_url = response["lighthouseResult"]["finalUrl"]

            is_first = i == 0
            if is_first:
                # Create and set up the workbook on the first iteration.
                print("Creating Excel workbook...")
                workbook = ExcelWorkbook(
                    final_url, metadata, audit_results, metrics_results
                )
                workbook.set_up_worksheet()
            else:
                # Simply update the workbook attrs after the first response.
                workbook.url = final_url
                workbook.metadata = metadata
                workbook.audit_results = audit_results
                workbook.metrics_results = metrics_results

            # Write the results to the Excel worksheet.
            workbook.write_to_worksheet(is_first)

        else:
            raise ValueError("Invalid format specified. Please try again.")

    if excel_output:
        # Calculate the avg score, close the workbook and save the Excel file.
        workbook.finalize_and_save()
